from datetime import timedelta
from django.utils import timezone
import dateutil.parser
import re


class ParseError(RuntimeError): 
    pass


class TextDateRange(object):
    """
    >>> TextDateRange('2008-04-01').tuple()
    ('2008-04-01', '2008-04-01')
    
    >>> TextDateRange('2008-4-1').tuple()
    ('2008-04-01', '2008-04-01')

    >>> TextDateRange('2008-04-01').num_days()
    1

    >>> TextDateRange('2008-04-01:2008-04-03').tuple()
    ('2008-04-01', '2008-04-03')

    >>> TextDateRange('2008-04-01:2008-04-03').num_days()
    3
    
    >>> TextDateRange('2007-12-29:2008-01-02').days_list()
    ['2007-12-29', '2007-12-30', '2007-12-31', '2008-01-01', '2008-01-02']

    >>> TextDateRange('2008-04-01:2008-04-03', set_end_date_time=True).tuple()
    ('2008-04-01', '2008-04-03 23:59:59')

    >>> TextDateRange('2008-04-01:2008-04-03', set_end_date_time=True).num_days()
    3

    >>> TextDateRange('2007-12-29:2008-01-02', set_end_date_time=True).days_list()
    ['2007-12-29', '2007-12-30', '2007-12-31', '2008-01-01', '2008-01-02']

    >>> TextDateRange('abc')
    Traceback (most recent call last):
      ...
    ParseError: Unable to parse date "abc"

    >>> TextDateRange('2008-04-03:2008-03-01')
    Traceback (most recent call last):
      ...
    ParseError: Enddate must be after the startdate
    """
    def __init__(self, instr, set_end_date_time=False):
        tokens = instr.split(':')
        try:
            date_strings = [self._token_to_date_string(x) for x in tokens]
            if len(date_strings) == 1:
                self.start, self.end = date_strings[0], date_strings[0]
            elif len(date_strings) == 2:
                self.start, self.end = date_strings[0], date_strings[1]
            else:
                raise ParseError
        except ParseError:
            raise ParseError('Unable to parse date "%s"' % instr)
        if self.start > self.end:
            raise ParseError('Enddate must be after the startdate')
        if set_end_date_time:
            self.end += ' 23:59:59'
    
    def _token_to_date_string(self, token):
        daysagoregex = re.compile(r'(\d+) ?days ?ago')
        dateregex = re.compile('^(\d{4})-(\d{1,2})-(\d{1,2})$')
        if token.lower() == 'yesterday':
            return self._format_datetime(timezone.now() - timedelta(days=1))
        elif token.lower() == 'today':
            return self._format_datetime(timezone.now())
        elif daysagoregex.match(token.lower()):
            days_ago = int(daysagoregex.match(token.lower()).groups()[0])
            return self._format_datetime(timezone.now() - timedelta(days=days_ago))
        elif dateregex.match(token):
            groups = dateregex.match(token).groups()
            return "%04d-%02d-%02d" % (int(groups[0]), int(groups[1]), int(groups[2]))
        else:
            raise ParseError
        
    def _set_start_end_from_days_ago(self, days_ago):
            date = self._format_datetime(timezone.now() - timedelta(days=days_ago))
            self.start, self.end = date, date
        
    def days_list(self):
        format_str = '%Y-%m-%d'
        date = timezone.now().strptime(self.start, format_str)
        enddate = timezone.now().strptime(_strip_time(self.end), format_str)
        increment_by = timedelta(days=1)
        output = []
        while date <= enddate:
            output.append(self._format_datetime(date))
            date += increment_by
        return output
    
    def tuple(self):
        return self.start, self.end

    def num_days(self):
        d1 = timezone.now().strptime(_strip_time(self.end), '%Y-%m-%d')
        d2 = timezone.now().strptime(self.start, '%Y-%m-%d')
        return (d1 - d2).days + 1

    @property
    def start_dateobj(self):
        return dateutil.parser.parse(self.start)

    @property
    def end_dateobj(self):
        return dateutil.parser.parse(self.end)
    
    @staticmethod
    def _format_datetime(datetime):
        format_str = '%Y-%m-%d'
        return datetime.strftime(format_str)

    def __str__(self):
        return "{0} -> {1}".format(self.start, self.end)


def _strip_time(date_with_time):
    return date_with_time.split(' ')[0]
        

if __name__ == '__main__':
    import doctest
    print("Running doctest . . .")
    doctest.testmod()


"""
This file demonstrates two different styles of tests (one doctest and one
unittest). These will both pass when you run "manage.py test".

Replace these with more appropriate tests for your application.
"""

from django.test import TestCase


class SimpleTest(TestCase):
    def test_basic_addition(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        self.assertEqual(1 + 1, 2)

__test__ = {"doctest": """
Another way to test that 1 + 1 is equal to 2.

>>> 1 + 1 == 2
True
"""}




from .reports import Report, Column, columns, filters, Filter, groupbys, GroupBy, DateFilter, \
    CalculatedColumn, PercentageCalculatedColumn, BooleanFilter, DecimalColumn, HiddenColumn, \
    ChoicesColumn, PercentageDifferenceColumn
from .reporting_site import site


REPORTING_SOURCE_FILE = 'reports'
default_app_config = 'django_spreadsheet_reports.autodiscover.MyAppConfig'



from django.db import models


class Bookmark(models.Model):
    name = models.CharField(max_length=255)
    uri = models.CharField(max_length=4096)

    class Meta:
        ordering = ('name',)

    def __unicode__(self):
        return self.name


class Notice(models.Model):
    slug = models.CharField(max_length=255)
    date = models.DateField(auto_now_add=True)
    creation_date = models.DateTimeField(auto_now_add=True)
    group_name = models.CharField(max_length=255)
    group_val = models.IntegerField(default=0)
    low_limit = models.IntegerField(default=0)
    high_limit = models.IntegerField(default=0)
    prev_val = models.IntegerField(default=0)

    def get_change_pct(self):
        before_data = float(self.prev_val)
        after_data = float(self.group_val)
        delta = after_data - before_data
        if before_data == 0.0:
            val = after_data
        else:
            val = delta / before_data
        val = val * 100.0
        return "%.0f" % val



from django.conf.urls import patterns, url
from django.contrib.admin.views.decorators import staff_member_required
from django.template import RequestContext
from django.shortcuts import render_to_response

from . import models


@staff_member_required
def home(request):
    context = RequestContext(request)
    context['available_reports'] = site._registered_reports
    context['bookmarks'] = models.Bookmark.objects.all()
    context['daily_reports'] = site._registered_daily_reports
    context['weekly_reports'] = site._registered_weekly_reports
    return render_to_response('django_spreadsheet_reports/home.html', context_instance=context)


def wrap(report):
    def wrapper(request, *args, **kwargs):
        return report.render_to_response(request)

    wrapper = staff_member_required(wrapper)
    return wrapper


class ReportingSite(object):
    _registered_reports = []
    _registered_daily_reports = []
    _registered_weekly_reports = []

    def register(self, report):
        if isinstance(report, type):
            report = report()
        self._registered_reports.append(report)

    def register_daily(self, report):
        if isinstance(report, type):
            report = report()
        self._registered_daily_reports.append(report)

    def get_daily_reports(self):
        return self._registered_daily_reports

    def register_weekly(self, report):
        if isinstance(report, type):
            report = report()
        self._registered_weekly_reports.append(report)

    def get_weekly_reports(self):
        return self._registered_weekly_reports

    def urls(self):
        urlpatterns = patterns('', url(r'^$', home, name='django_spreadsheet_reports'))
        for report in self._registered_reports:
            urlpatterns += patterns('', url(r'^%s/$' % report.slug, wrap(report),
                                            name='django_spreadsheet_reports-%s' % report.slug))
        for report in self._registered_daily_reports:
            urlpatterns += patterns('', url(r'^%s/$' % report.slug, wrap(report),
                                            name='django_spreadsheet_reports-%s' % report.slug))
        for report in self._registered_weekly_reports:
            urlpatterns += patterns('', url(r'^%s/$' % report.slug, wrap(report),
                                            name='django_spreadsheet_reports-%s' % report.slug))
        return urlpatterns

    urls = property(urls)
site = ReportingSite()



import importlib.util
from django.apps.config import AppConfig
from django.conf import settings

def autodiscover():
    # Check django/contrib/admin/__init__.py to know what I'm doing :)
    for app in settings.INSTALLED_APPS:
        reports_spec = app + ".reports"
        if importlib.util.find_spec(reports_spec) is None:
            continue

        __import__(reports_spec)


class MyAppConfig(AppConfig):
    name = 'django_spreadsheet_reports'

    def ready(self):
        autodiscover()




import logging
import datetime

from django.db.models import Sum
from django.core.mail import send_mail
from django.conf import settings
from django.template.loader import render_to_string

import django_spreadsheet_reports
from django_spreadsheet_reports.models import Notice


logger = logging.getLogger(__name__)


def save_reports(report_type='daily', should_send_mail=True):
    logger.info("Getting %s reports" % report_type)
    if report_type == 'daily':
        cutoff = datetime.date.today() - datetime.timedelta(days=2)
        reports = django_spreadsheet_reports.site.get_daily_reports()
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
    elif report_type == 'weekly':
        cutoff = datetime.date.today() - datetime.timedelta(days=8)
        reports = django_spreadsheet_reports.site.get_weekly_reports()
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
        two_weeks = datetime.date.today() - datetime.timedelta(days=15)
    else:
        logger.error("%s is not a valid report type" % report_type)
        raise RuntimeError()
    count = 0
    for report in reports:
        logger.info("Getting stats for report %s" % report.name)
        qs = report.notice_model.objects.filter(**report.notices_filter_by)
        if report_type == 'daily':            
            qs = qs.filter(date=yesterday).values(
                report.notice_group_by).annotate(Sum(report.notice_field))
        elif report_type == 'weekly':
            qs = qs.filter(date__lte=yesterday, date__gte=cutoff).values(
                report.notice_group_by).annotate(Sum(report.notice_field))

        for row in qs:
            high_limit = low_limit = 0
            group_by = row[report.notice_group_by]
            val = float(row[report.notice_field+'__sum'])
            if report_type == 'daily':
                filters = {report.notice_group_by: group_by,
                           'date': cutoff}
            elif report_type == 'weekly':
                # noinspection PyUnboundLocalVariable
                filters = {report.notice_group_by: group_by,
                           'date__lte': cutoff,
                           'date__gte': two_weeks}
            else:
                raise RuntimeError()
            old_mark = report.notice_model.objects.filter(
                **report.notices_filter_by).filter(
                **filters).values(report.notice_group_by).annotate(
                Sum(report.notice_field))
        
            is_visible = True
            prev_val = 0.0
            if old_mark:
                mark = old_mark[0]
                prev_val = float(mark[report.notice_field+'__sum'])
                logger.info("Got prev_val of %s" % prev_val)
                high_limit = prev_val + (prev_val * report.change_threshold)
                low_limit = prev_val - (prev_val * report.change_threshold)
                if low_limit <= val <= high_limit:
                    is_visible = False
            if val < report.minimum_threshold and \
               prev_val < report.minimum_threshold:
                is_visible = False
                
            if is_visible:
                Notice.objects.create(slug=report.slug,
                                      group_name=group_by,
                                      group_val=val,
                                      prev_val=prev_val,
                                      low_limit=low_limit,
                                      high_limit=high_limit)
            logger.info("Created notice for %s" % group_by)
            count += 1
        logger.info("Created %s notices" % count)


def send_new_notices_report():
    logger.info("sending email to staff members")
    subject = 'New notices report'
    cutoff = datetime.datetime.now() - datetime.timedelta(hours=24)
    notices = Notice.objects.filter(creation_date__gte=cutoff)
    staff_members = []  # TODO make this a list of staff usernames
    message = render_to_string('django_spreadsheet_reports/notice_list.txt',
                               {'notices': notices})
    logger.info("sending emails to %s staff members" % len(staff_members))
    send_mail(subject, message, settings.DEFAULT_FROM_EMAIL, staff_members,
              fail_silently=False)
    logger.info("finished")


def load_from_urls():
    from django.conf import settings
    __import__(settings.ROOT_URLCONF)
